import base64
import datetime
import re
import typing

import orjson

from ._dates import parse_timestamp
from ._regex import (
    RE_DOMAIN_NAMES,
    RE_EMAIL,
    RE_HTML,
    RE_JAVASCRIPT,
    RE_PHONE_NUMBER,
    RE_UUID,
    REGEX_CAMEL_CASE,
)


def is_uuid(s: str) -> bool:
    return (
        re.match(r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$", s)
        is not None
    )


def camel_to_slug(s: str, force_lower: bool = True) -> str:
    # Step 1: Handle acronyms and numbers, keeping numbers joined with preceding letters
    s = REGEX_CAMEL_CASE.sub(r"_", s)

    # Step 2: Remove any leading underscore
    s = s.lstrip("_")

    # Step 3: Convert to lowercase and replace multiple underscores with single ones
    return re.sub(r"_+", "_", s.lower() if force_lower else s)

def camel_to_snake(name: str) -> str:
    return re.sub(r"(?<!^)(?=[A-Z])", "_", name).lower()

def camel_to_kebab(name: str) -> str:
    return re.sub(r"(?<!^)(?=[A-Z])", "-", name).lower()

def snake_to_camel(name: str) -> str:
    return "".join(word.title() for word in name.split("_"))

def snake_to_kebab(name: str) -> str:
    return name.replace("_", "-")

def kebab_to_camel(name: str) -> str:
    return "".join(word.title() for word in name.split("-"))

def kebab_to_snake(name: str) -> str:
    return name.replace("-", "_")

def slug_to_camel(s: str) -> str:
    return "".join(word.title() for word in s.split("_"))

def slug_to_snake(s: str) -> str:
    return s.replace("-", "_")

def slug_to_kebab(s: str) -> str:
    return s.replace("_", "-")


def detect_string_type(
    string: str,
) -> typing.Literal[
    "url-image",
    "url-video",
    "url",
    "uuid",
    "base64-encoded-string",
    "email-address",
    "html",
    "domain-name",
    "phone-number",
    "json-string",
    "javascript-snippet",
    "date-string",
    "timestamp-string",
    "unknown",
]:
    # Check for URLs
    if re.match(r"^https?://", string):
        if re.match(r"\.(jpeg|jpg|png|gif|bmp)$", string):
            return "url-image"
        elif re.match(r"\.(mp4|webm|ogv)$", string):
            return "url-video"
        else:
            return "url"

    # Check for UUIDs
    if RE_UUID.match(string):
        return "uuid"

    # # Check for base64-encoded strings
    try:
        if base64.b64encode(base64.b64decode(string)) == string:
            return "base64-encoded-string"
    except Exception:
        pass
    # if re.match(r'^(?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?$', string):
    #   return 'base64-encoded-string'

    # Check for email addresses
    if RE_EMAIL.match(string):
        return "email-address"

    if RE_HTML.match(string):
        return "html"

    # Check for domain names
    if RE_DOMAIN_NAMES.match(string):
        return "domain-name"

    # Check for phone numbers
    if RE_PHONE_NUMBER.match(string):
        return "phone-number"

    # Check for JSON strings
    try:
        orjson.loads(string)
        return "json-string"
    except ValueError:
        pass

    # Check for JavaScript snippets
    if RE_JAVASCRIPT.match(string):
        return "javascript-snippet"

    # Check for date and timestamp strings
    try:
        datetime.datetime.strptime(string, "%Y-%m-%d")
        return "date-string"
    except ValueError:
        pass
    try:
        parse_timestamp(string, raise_error=True)
        return "timestamp-string"
    except Exception:
        pass

    return "unknown"



import pytest
import asyncio
from unittest.mock import patch, MagicMock
from good_common.utilities._asyncio import run_async


def test_run_async_with_existing_loop():
    def sample_function():
        return "Hello, World!"

    with patch("asyncio.get_event_loop") as mock_get_loop:
        mock_loop = MagicMock()
        mock_loop.run_until_complete.return_value = "Hello, World!"
        mock_get_loop.return_value = mock_loop

        result = run_async(sample_function())

        mock_get_loop.assert_called_once()
        mock_loop.run_until_complete.assert_called_once()
        assert result == "Hello, World!"


def test_run_async_without_existing_loop():
    def sample_function():
        return "Hello, World!"

    with patch(
        "asyncio.get_event_loop", side_effect=[RuntimeError, MagicMock()]
    ), patch("asyncio.new_event_loop") as mock_new_loop, patch(
        "asyncio.set_event_loop"
    ) as mock_set_loop:
        mock_loop = MagicMock()
        mock_loop.run_until_complete.return_value = "Hello, World!"
        mock_new_loop.return_value = mock_loop

        result = run_async(sample_function())

        mock_new_loop.assert_called_once()
        mock_set_loop.assert_called_once_with(mock_loop)
        mock_loop.run_until_complete.assert_called_once()
        assert result == "Hello, World!"


def test_run_async_in_jupyter():
    def sample_function():
        return "Hello, Jupyter!"

    with patch.dict("sys.modules", {"IPython": MagicMock()}), patch(
        "nest_asyncio.apply"
    ) as mock_apply, patch("asyncio.get_event_loop") as mock_get_loop:
        mock_loop = MagicMock()
        mock_loop.run_until_complete.return_value = "Hello, Jupyter!"
        mock_get_loop.return_value = mock_loop

        result = run_async(sample_function())

        mock_apply.assert_called_once()
        assert result == "Hello, Jupyter!"


def test_run_async_with_exception():
    def sample_function():
        raise ValueError("Test exception")

    with patch("asyncio.get_event_loop") as mock_get_loop:
        mock_loop = MagicMock()
        mock_loop.run_until_complete.side_effect = ValueError("Test exception")
        mock_get_loop.return_value = mock_loop

        with pytest.raises(ValueError, match="Test exception"):
            run_async(sample_function())


def test_run_async_with_real_coroutine():
    async def sample_coroutine():
        await asyncio.sleep(0.1)
        return "Hello, Async World!"

    # Create a new event loop for this test
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    try:
        result = run_async(sample_coroutine())
        assert result == "Hello, Async World!"
    finally:
        # Clean up the event loop
        loop.close()



import typing
import re
from collections.abc import Mapping, Sequence


T = typing.TypeVar("T")


def try_chain(
    fns: list[typing.Callable[..., T]], fail=False, default_value: T | None = None
) -> typing.Callable[..., T | None]:
    def _try_chain(value):
        for fn in fns:
            try:
                return fn(value)
            except Exception:
                continue
        if fail:
            raise ValueError(f"Could not process {value}")
        return default_value

    return _try_chain


T = typing.TypeVar("T")
X = typing.TypeVar("X")


def deep_attribute_get(
    obj: typing.Any,
    path: str,
    default: typing.Any = None,
    debug: bool = False,
    return_paths: bool = False,
) -> typing.Union[
    typing.Any, typing.List[typing.Any], typing.List[typing.Tuple[typing.Any, str]]
]:
    segments = re.split(r"\.(?![^\[]*\])", path)
    if debug:
        print(f"Path segments: {segments}")
    result = _traverse(obj, segments, default, debug, "", return_paths)
    flattened = _flatten_and_filter(result, return_paths)

    if return_paths:
        filtered = [(value, path) for value, path in flattened if value is not None]
        return filtered[0] if len(filtered) == 1 else filtered
    else:
        filtered = [value for value in flattened if value is not None]
        return filtered[0] if len(filtered) == 1 else filtered


def _traverse(
    obj: typing.Any,
    segments: list[str],
    default: typing.Any,
    debug: bool,
    current_path: str,
    return_paths: bool,
) -> typing.Any:
    if not segments:
        return (obj, current_path) if return_paths else obj

    current_segment = segments[0]
    remaining_segments = segments[1:]

    if debug:
        print(f"Current segment: {current_segment}")
        print(f"Object type: {type(obj)}")

    if current_segment == "*":
        if isinstance(obj, Mapping):
            return [
                _traverse(
                    value,
                    remaining_segments,
                    default,
                    debug,
                    f"{current_path}.{key}",
                    return_paths,
                )
                for key, value in obj.items()
            ]
        elif isinstance(obj, Sequence) and not isinstance(obj, (str, bytes)):
            return [
                _traverse(
                    item,
                    remaining_segments,
                    default,
                    debug,
                    f"{current_path}[{i}]",
                    return_paths,
                )
                for i, item in enumerate(obj)
            ]
        else:
            if debug:
                print(f"Wildcard '*' not applicable for object of type {type(obj)}")
            return default

    if "[" in current_segment and "]" in current_segment:
        key, index = current_segment.split("[", 1)
        index = index.rstrip("]")
        if key:
            obj = anyget(obj, key, default)
            current_path = f"{current_path}.{key}" if current_path else key
            if obj is default:
                if debug:
                    print(f"Key '{key}' not found in object")
                return default
        if index == "*":
            if isinstance(obj, Sequence) and not isinstance(obj, (str, bytes)):
                return [
                    _traverse(
                        item,
                        remaining_segments,
                        default,
                        debug,
                        f"{current_path}[{i}]",
                        return_paths,
                    )
                    for i, item in enumerate(obj)
                ]
            else:
                if debug:
                    print(
                        f"Wildcard '[*]' not applicable for object of type {type(obj)}"
                    )
                return default
        try:
            obj = obj[int(index)]
            current_path = f"{current_path}[{index}]"
        except (IndexError, ValueError, TypeError):
            if debug:
                print(f"Invalid index {index} for object of type {type(obj)}")
            return default
    else:
        if isinstance(obj, Mapping):
            matching_keys = [
                k for k in obj.keys() if re.fullmatch(current_segment, str(k))
            ]
            if debug:
                print(f"Matching keys for pattern '{current_segment}': {matching_keys}")
            if len(matching_keys) == 1:
                obj = obj[matching_keys[0]]
                current_path = (
                    f"{current_path}.{matching_keys[0]}"
                    if current_path
                    else matching_keys[0]
                )
            elif len(matching_keys) > 1:
                return [
                    _traverse(
                        obj[k],
                        remaining_segments,
                        default,
                        debug,
                        f"{current_path}.{k}",
                        return_paths,
                    )
                    for k in matching_keys
                ]
            else:
                if debug:
                    print(f"No matching keys found for pattern '{current_segment}'")
                return default
        # if isinstance(obj, Mapping):
        #     matching_keys = [k for k in obj.keys() if re.match(current_segment, str(k))]
        #     if debug:
        #         print(f"Matching keys for pattern '{current_segment}': {matching_keys}")
        #     if len(matching_keys) == 1:
        #         obj = obj[matching_keys[0]]
        #         current_path = (
        #             f"{current_path}.{matching_keys[0]}"
        #             if current_path
        #             else matching_keys[0]
        #         )
        #     elif len(matching_keys) > 1:
        #         return [
        #             _traverse(
        #                 obj[k],
        #                 remaining_segments,
        #                 default,
        #                 debug,
        #                 f"{current_path}.{k}",
        #                 return_paths,
        #             )
        #             for k in matching_keys
        #         ]
        #     else:
        #         if debug:
        #             print(f"No matching keys found for pattern '{current_segment}'")
        #         return default
        else:
            obj = anyget(obj, current_segment, default)
            current_path = (
                f"{current_path}.{current_segment}" if current_path else current_segment
            )

        if obj is default:
            if debug:
                print(f"Segment '{current_segment}' not found in object")
            return default

    return _traverse(
        obj, remaining_segments, default, debug, current_path, return_paths
    )


def _flatten_and_filter(result: typing.Any, return_paths: bool) -> typing.List:
    if isinstance(result, list):
        return [
            item
            for sublist in [
                _flatten_and_filter(i, return_paths) for i in result if i is not None
            ]
            for item in sublist
        ]
    if return_paths:
        return [result] if result[0] is not None else []
    return [result] if result is not None else []


def anyget(obj: typing.Any, key: str, default: X = None) -> T | X:
    if isinstance(obj, Mapping):
        return obj.get(key, default)
    elif hasattr(obj, key):
        return getattr(obj, key)
    return default


def anyset(obj: dict[str, T], key, value: T) -> dict[str, T]:
    if isinstance(obj, dict):
        obj[key] = value
    else:
        setattr(obj, key, value)
    return obj


def deep_attribute_set(obj: typing.Any, path: str, value: typing.Any) -> typing.Any:
    if "." not in path:
        return anyset(obj, path, value)
    key, rest = path.split(".", 1)
    return deep_attribute_set(anyget(obj, key, {}), rest, value)


def set_defaults(__base: dict | None = None, **kwargs):
    if not __base:
        __base = {}
    for k, v in kwargs.items():
        if k not in __base:
            __base[k] = v
    return __base


# def filter_nulls(obj):
#     if isinstance(obj, dict):
#         return {k: filter_nulls(v) for k, v in obj.items() if v}
#     elif isinstance(obj, list):
#         return [filter_nulls(v) for v in obj if v]
#     else:
#         return obj
def filter_nulls(obj):
    if isinstance(obj, dict):
        return {
            k: v
            for k, v in ((k, filter_nulls(v)) for k, v in obj.items())
            if v not in (None, {}, [])
        }
    elif isinstance(obj, list):
        return [v for v in (filter_nulls(v) for v in obj) if v not in (None, {}, [])]
    else:
        return obj



import pytest
import datetime
from good_common.utilities._iterators import iter_ranges


def test_date_ranges_with_int_step():
    start = datetime.date(2023, 1, 1)
    end = datetime.date(2023, 1, 5)
    result = list(iter_ranges(start, end, 2))
    expected = [
        (datetime.date(2023, 1, 1), datetime.date(2023, 1, 3)),
        (datetime.date(2023, 1, 3), datetime.date(2023, 1, 5)),
    ]
    assert result == expected


def test_datetime_ranges_with_int_step():
    start = datetime.datetime(2023, 1, 1, 12, 0)
    end = datetime.datetime(2023, 1, 3, 12, 0)
    result = list(iter_ranges(start, end, 1))
    expected = [
        (datetime.datetime(2023, 1, 1, 12, 0), datetime.datetime(2023, 1, 2, 12, 0)),
        (datetime.datetime(2023, 1, 2, 12, 0), datetime.datetime(2023, 1, 3, 12, 0)),
    ]
    assert result == expected


def test_date_ranges_with_timedelta_step():
    start = datetime.date(2023, 1, 1)
    end = datetime.date(2023, 1, 5)
    step = datetime.timedelta(days=2)
    result = list(iter_ranges(start, end, step))
    expected = [
        (datetime.date(2023, 1, 1), datetime.date(2023, 1, 3)),
        (datetime.date(2023, 1, 3), datetime.date(2023, 1, 5)),
    ]
    assert result == expected


def test_single_day_range():
    start = datetime.date(2023, 1, 1)
    end = datetime.date(2023, 1, 2)
    result = list(iter_ranges(start, end))
    expected = [(datetime.date(2023, 1, 1), datetime.date(2023, 1, 2))]
    assert result == expected


def test_empty_range():
    start = datetime.date(2023, 1, 1)
    end = datetime.date(2023, 1, 1)
    result = list(iter_ranges(start, end))
    assert result == []


def test_reversed_range():
    start = datetime.date(2023, 1, 5)
    end = datetime.date(2023, 1, 1)
    result = list(iter_ranges(start, end))
    assert result == []


def test_large_step():
    start = datetime.date(2023, 1, 1)
    end = datetime.date(2023, 1, 5)
    result = list(iter_ranges(start, end, 10))
    expected = [(datetime.date(2023, 1, 1), datetime.date(2023, 1, 5))]
    assert result == expected


def test_datetime_with_timedelta():
    start = datetime.datetime(2023, 1, 1, 12, 0)
    end = datetime.datetime(2023, 1, 3, 12, 0)
    step = datetime.timedelta(hours=24)
    result = list(iter_ranges(start, end, step))
    expected = [
        (datetime.datetime(2023, 1, 1, 12, 0), datetime.datetime(2023, 1, 2, 12, 0)),
        (datetime.datetime(2023, 1, 2, 12, 0), datetime.datetime(2023, 1, 3, 12, 0)),
    ]
    assert result == expected



import typing
from pathlib import Path

import yaml
# from yaml import dump, load

try:
    from yaml import CDumper as Dumper
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Dumper, Loader


#
# YAML Functions
#


def str_presenter(dumper, data):
    text_list = [line.rstrip() for line in data.splitlines()]
    fixed_data = "\n".join(text_list)
    if len(text_list) > 1:
        return dumper.represent_scalar("tag:yaml.org,2002:str", fixed_data, style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", fixed_data)


Dumper.add_representer(str, str_presenter)
yaml.representer.SafeRepresenter.add_representer(
    str, str_presenter
)  # to use with safe_dum


def yaml_load(path) -> typing.Any:
    with open(path, "r") as f:
        return yaml.load(f, Loader=Loader)


def yaml_dumps(data: typing.Any, sort_keys: bool = False, **kwargs) -> str:
    return yaml.dump(data, Dumper=Dumper, sort_keys=sort_keys, **kwargs)


def yaml_loads(data: str) -> typing.Any:
    return yaml.load(data, Loader=Loader)


def yaml_dump(path: str | Path, data: typing.Any, sort_keys=False, **kwargs) -> None:
    with open(path, "w") as f:
        return yaml.dump(data, f, Dumper=Dumper, sort_keys=sort_keys, **kwargs)



from contextlib import contextmanager
from time import perf_counter

from loguru import logger


@contextmanager
def catchtime(name: str):
    """
    Derived from # https://stackoverflow.com/questions/33987060/python-context-manager-that-measures-time
    """
    logger.info(f"{name}: starting...")
    start = perf_counter()
    yield lambda: perf_counter() - start
    logger.info(f"\t\t complete in {perf_counter() - start:.3f} seconds")



import pytest
import asyncio
from typing import Annotated
from good_common.pipeline import Pipeline, Attribute, function_mapper


# Test data for the tests
@pytest.fixture
def basic_pipeline():
    def add(a: int, b: int) -> Annotated[int, Attribute("result")]:
        return a + b

    def multiply(result: int, factor: int) -> Annotated[int, Attribute("result")]:
        return result * factor

    return Pipeline(add, multiply)


@pytest.fixture
def async_pipeline():
    async def async_add(a: int, b: int) -> Annotated[int, Attribute("sum")]:
        await asyncio.sleep(0.1)
        return a + b

    async def async_multiply(
        sum: int, factor: int
    ) -> Annotated[int, Attribute("product")]:
        await asyncio.sleep(0.1)
        return sum * factor

    return Pipeline(async_add, async_multiply)


@pytest.fixture
def mixed_pipeline():
    async def async_add(a: int, b: int) -> Annotated[int, Attribute("sum")]:
        await asyncio.sleep(0.1)
        return a + b

    def sync_multiply(sum: int, factor: int) -> Annotated[int, Attribute("product")]:
        return sum * factor

    return Pipeline(async_add, sync_multiply)


# Test cases
@pytest.mark.asyncio
async def test_basic_sync_pipeline(basic_pipeline):
    output = await basic_pipeline(a=2, b=3, factor=4)
    assert output.result == 20


@pytest.mark.asyncio
async def test_async_pipeline(async_pipeline):
    output = await async_pipeline(a=2, b=3, factor=4)
    assert output.product == 20


@pytest.mark.asyncio
async def test_mixed_pipeline(mixed_pipeline):
    output = await mixed_pipeline(a=2, b=3, factor=4)
    assert output.product == 20


@pytest.mark.asyncio
async def test_pipeline_with_annotated_return_type():
    def create_greeting(name: str) -> Annotated[str, Attribute("greeting")]:
        return f"Hello, {name}!"

    # Wrap the print_greeting in a pipeline-compatible function
    def print_greeting(greeting: str) -> Annotated[None, Attribute("finished")]:
        print(f"Greeting: {greeting}")
        return None

    pipeline = Pipeline(create_greeting, print_greeting)
    output = await pipeline(name="Alice")
    assert output.greeting == "Hello, Alice!"


@pytest.mark.asyncio
async def test_pipeline_with_async_and_sync_mapped():
    async def async_subtract(x: int, y: int) -> Annotated[int, Attribute("diff")]:
        await asyncio.sleep(0.1)
        return x - y

    def multiply_diff(
        diff: int, factor: int
    ) -> Annotated[int, Attribute("final_result")]:
        return diff * factor

    # Fix the mapping and ensure 'diff' is correctly translated
    pipeline = Pipeline(async_subtract, function_mapper(multiply_diff, diff="diff"))

    output = await pipeline(x=10, y=3, factor=5)
    assert output.final_result == 35  # (10 - 3) * 5 = 35


@pytest.mark.asyncio
async def test_pipeline_parallel_execution():
    async def async_add(a: int, b: int) -> Annotated[int, Attribute("sum")]:
        await asyncio.sleep(0.1)
        return a + b

    def multiply(sum: int, factor: int) -> Annotated[int, Attribute("product")]:
        return sum * factor

    pipeline = Pipeline(async_add, multiply)

    inputs = [
        {"a": 1, "b": 2, "factor": 2},
        {"a": 2, "b": 3, "factor": 3},
        {"a": 3, "b": 4, "factor": 4},
    ]

    results = [
        result
        async for result in pipeline.execute(
            *inputs, max_workers=3, display_progress=False
        )
    ]

    assert results[0].unwrap().product == 6  # (1+2) * 2
    assert results[1].unwrap().product == 15  # (2+3) * 3
    assert results[2].unwrap().product == 28  # (3+4) * 4


@pytest.mark.asyncio
async def test_pipeline_with_errors():
    def faulty_add(a: int, b: int) -> Annotated[int, Attribute("result")]:
        raise ValueError("Deliberate error!")

    pipeline = Pipeline(faulty_add)

    with pytest.raises(ValueError, match="Deliberate error!"):
        await pipeline(a=1, b=2)


@pytest.mark.asyncio
async def test_pipeline_with_async_errors():
    async def async_faulty_add(a: int, b: int) -> Annotated[int, Attribute("result")]:
        await asyncio.sleep(0.1)
        raise ValueError("Deliberate async error!")

    pipeline = Pipeline(async_faulty_add)

    with pytest.raises(ValueError, match="Deliberate async error!"):
        await pipeline(a=1, b=2)


@pytest.mark.asyncio
async def test_pipeline_parallel_execution_with_errors():
    async def async_add(a: int, b: int) -> Annotated[int, Attribute("sum")]:
        await asyncio.sleep(0.1)
        if a == 2:
            raise ValueError("Error on purpose!")
        return a + b

    pipeline = Pipeline(async_add)

    inputs = [
        {"a": 1, "b": 2},
        {"a": 2, "b": 3},  # This one will raise an error
        {"a": 3, "b": 4},
    ]

    results = [
        result
        async for result in pipeline.execute(
            *inputs, max_workers=3, display_progress=False
        )
    ]

    assert results[0].is_ok() and results[0].unwrap().sum == 3
    assert results[1].is_err()
    assert results[2].is_ok() and results[2].unwrap().sum == 7



import datetime
from typing import Union, Tuple, Generator, overload

DateType = Union[datetime.date, datetime.datetime]
StepType = Union[int, datetime.timedelta]


@overload
def iter_ranges(
    start: datetime.date, end: datetime.date, step: int = 1
) -> Generator[Tuple[datetime.date, datetime.date], None, None]: ...


@overload
def iter_ranges(
    start: datetime.datetime, end: datetime.datetime, step: int = 1
) -> Generator[Tuple[datetime.datetime, datetime.datetime], None, None]: ...


@overload
def iter_ranges(
    start: datetime.date, end: datetime.date, step: datetime.timedelta
) -> Generator[Tuple[datetime.date, datetime.date], None, None]: ...


@overload
def iter_ranges(
    start: datetime.datetime, end: datetime.datetime, step: datetime.timedelta
) -> Generator[Tuple[datetime.datetime, datetime.datetime], None, None]: ...


def iter_ranges(
    start: DateType, end: DateType, step: StepType = 1
) -> Generator[Tuple[DateType, DateType], None, None]:
    current = start
    if isinstance(step, int):
        delta = datetime.timedelta(days=step)
    else:
        delta = step

    while current < end:
        next_date = min(current + delta, end)
        yield current, next_date
        current = next_date
        if current == end:
            break



import pytest
from fast_depends import inject
from good_common.dependencies import BaseProvider, AsyncBaseProvider


def test__basic_base_provider():
    class FakeClient:
        def __init__(self, host: str, port: int):
            self.host = host
            self.port = port

        def __str__(self):
            return f"{self.host}:{self.port}"

    class ClientProvider(BaseProvider[FakeClient], FakeClient):
        pass

    @inject
    def test_client(client: FakeClient = ClientProvider(host="localhost", port=8080)):
        assert client.host == "localhost"
        assert client.port == 8080

    test_client()


def test__dependency_with_runtime_config():
    class FakeClient:
        def __init__(self, host: str, port: int, db: str):
            self.host = host
            self.port = port
            self.db = db

        def __str__(self):
            return f"{self.host}:{self.port}:{self.db}"

    class ClientProvider(BaseProvider[FakeClient], FakeClient):
        def initializer(self, cls_args: tuple, cls_kwargs: dict, fn_kwargs: dict):
            if fn_kwargs.get("db"):
                cls_kwargs["db"] = fn_kwargs["db"]
            return cls_args, cls_kwargs

    @inject
    def test_client(
        db: str,
        client: FakeClient = ClientProvider(host="localhost", port=8080),
    ):
        assert client.host == "localhost"
        assert client.port == 8080
        assert client.db == "test"

    test_client(db="test")


@pytest.mark.asyncio
async def test__basic_async_base_provider():
    class FakeClient:
        def __init__(self, host: str, port: int):
            self.host = host
            self.port = port

        def __str__(self):
            return f"{self.host}:{self.port}"

    class ClientProvider(AsyncBaseProvider[FakeClient], FakeClient):
        pass

    @inject
    async def test_client(
        client: FakeClient = ClientProvider(host="localhost", port=8080),
    ):
        assert client.host == "localhost"
        assert client.port == 8080

    await test_client()



import copy
import typing

from fast_depends.library import CustomField
from loguru import logger

T = typing.TypeVar("T")


class BaseProvider(CustomField, typing.Generic[T]):
    __override_class__: typing.ClassVar = None

    @classmethod
    def provide(cls, *args, **kwargs) -> T:
        if cls.__override_class__ is not None:
            return cls.__override_class__(*args, **kwargs)
        else:
            return cls.__bases__[1](*args, **kwargs)

    def __init__(self, *args, _debug: bool = False, **kwargs):
        super().__init__(cast=False)
        self._args = args
        self._kwargs = kwargs
        self._debug = _debug

    def initializer(
        self,
        cls_args: typing.Tuple[typing.Any, ...],
        cls_kwargs: typing.Dict[str, typing.Any],
        fn_kwargs: typing.Dict[str, typing.Any],
    ):
        return cls_args, cls_kwargs

    def use(self, /, **kwargs: typing.Dict[str, typing.Any]):
        if self._debug:
            logger.debug(f"Using {self.__class__.__name__}: {kwargs}")
        kwargs = super().use(**kwargs)
        if self.param_name:
            _args, _kwargs = self.initializer(
                cls_args=copy.copy(self._args),
                cls_kwargs=copy.copy(self._kwargs),
                fn_kwargs=kwargs,
            )
            kwargs[self.param_name] = self.provide(*_args, **_kwargs)  # type: ignore
        return kwargs

    def get(self, **kwargs) -> T:
        self.param_name = "_default"
        return self.use(**kwargs).get(self.param_name)  # type: ignore


class AsyncBaseProvider(CustomField, typing.Generic[T]):
    __override_class__: typing.ClassVar = None

    @classmethod
    async def provide(cls, *args, **kwargs) -> T:
        if cls.__override_class__ is not None:
            return cls.__override_class__(*args, **kwargs)
        else:
            return cls.__bases__[1](*args, **kwargs)

    def __init__(self, *args, _debug: bool = False, **kwargs):
        super().__init__(cast=False)
        self._args = args
        self._kwargs = kwargs
        self._debug = _debug

    def initializer(
        self,
        cls_args: typing.Tuple[typing.Any, ...],
        cls_kwargs: typing.Dict[str, typing.Any],
        fn_kwargs: typing.Dict[str, typing.Any],
    ):
        return cls_args, cls_kwargs

    async def use(self, /, **kwargs: typing.Dict[str, typing.Any]):
        if self._debug:
            logger.debug(f"Using {self.__class__.__name__}: {kwargs}")
        kwargs = super().use(**kwargs)
        if self.param_name:
            _args, _kwargs = self.initializer(
                cls_args=copy.copy(self._args),
                cls_kwargs=copy.copy(self._kwargs),
                fn_kwargs=kwargs,
            )
            kwargs[self.param_name] = await self.provide(*_args, **_kwargs)  # type: ignore
        return kwargs

    async def get(self, **kwargs) -> T:
        self.param_name = "_default"
        return await self.use(**kwargs).get(self.param_name)  # type: ignore



from __future__ import annotations

import collections
import dataclasses
import glob
import typing
from collections.abc import Iterable, MutableMapping

import farmhash
import jsonlines
import orjson
import slugify
from box import Box
from pydantic import BaseModel
from pydantic.main import IncEx

# Type aliases
DictLike = typing.Union[dict, "DotDict"]
D = typing.TypeVar("D", bound=dict | MutableMapping)
TKey = str | int
TKeys = tuple[TKey, ...]
TVal = str | int | float | bool | None

# Dictionary operations

T = typing.TypeVar("T", bound=dict | list)

def sort_object_keys(obj: T) -> T:
    if isinstance(obj, dict):
        return {k: sort_object_keys(v) for k, v in sorted(obj.items())}
    if isinstance(obj, list):
        return [sort_object_keys(i) for i in obj]
    return obj


def object_farmhash(obj: BaseModel | typing.Any, exclude_keys: IncEx = None) -> int:
    """
    Generate a consistent FarmHash64 hash from an object (include Pydantic models)

    :param obj: A Pydantic model instance
    :return: 64-bit FarmHash as an integer
    """

    def serialize(data: typing.Any) -> typing.Any:
        if isinstance(data, BaseModel):
            return serialize(
                sort_object_keys(
                    data.model_dump(
                        mode="json",
                        exclude_none=True,
                        exclude=exclude_keys
                    )
                )
            )
        elif isinstance(data, dict):
            return {k: serialize(v) for k, v in sorted(data.items())}
        elif isinstance(data, list):
            return [serialize(item) for item in sorted(data)]
        elif isinstance(data, tuple):
            return tuple(serialize(item) for item in data)
        elif isinstance(data, set):
            return frozenset(serialize(item) for item in sorted(data))
        else:
            return str(data)

    serialized_data = serialize(obj)
    data_str = str(serialized_data)
    return farmhash.hash64(data_str)


def merge_dicts(
    dicts: list[dict],
    keep_unique_values: bool = False,
    keep_unique_value_keys: list = [],
) -> dict:
    """
    Merge a list of dictionaries into a single dictionary.

    Args:
        dicts (list[dict]): List of dictionaries to merge.
        keep_unique_values (bool, optional): If True, keep all unique values for each key. Defaults to False.
        keep_unique_value_keys (list, optional): List of keys for which to keep unique values. Defaults to [].

    Returns:
        dict: Merged dictionary.
    """
    merged: typing.Dict = {}
    for d in deduplicate_dicts(dicts):
        for key, value in d.items():
            if isinstance(value, dict):
                if key in merged:
                    if isinstance(merged[key], list):
                        merged[key] = {k: v for k, v in enumerate(merged[key])}
                    merged_dict = merge_dicts(
                        [merged[key], value],
                        keep_unique_values=keep_unique_values,
                        keep_unique_value_keys=keep_unique_value_keys,
                    )
                    merged[key] = merged_dict
                else:
                    merged[key] = value
            elif isinstance(value, list):
                if key in merged:
                    for elem in value:
                        if elem not in merged[key]:
                            merged[key].append(elem)
                else:
                    merged[key] = value
            else:
                if value is not None and value != "":
                    if (
                        keep_unique_values and key in merged
                    ) or key in keep_unique_value_keys:
                        if not isinstance(merged[key], list):
                            merged[key] = [merged[key]]

                        if value not in merged[key]:
                            merged[key].append(value)

                        if len(merged[key]) == 1:
                            merged[key] = merged[key][0]
                    else:
                        merged[key] = value

    return {k: v for k, v in merged.items() if v is not None and v != ""}


def deduplicate_dicts(dicts: list[dict]) -> list[dict]:
    """
    Remove duplicate dictionaries from a list of dictionaries.

    Args:
        dicts (list[dict]): List of dictionaries to deduplicate.

    Returns:
        list[dict]: List of unique dictionaries.
    """
    new_dicts = []
    for d in dicts:
        if d not in new_dicts:
            new_dicts.append(d)
    return new_dicts


def clean_dict(d: typing.Union[dict, list, set]) -> typing.Union[dict, list]:
    """
    Convert defaultdict to plain dictionary recursively and handle unhashable types.

    Args:
        d (Union[dict, list, set]): Input data structure to clean.

    Returns:
        Union[dict, list]: Cleaned data structure.
    """
    if isinstance(d, dict):
        return {k: clean_dict(v) for k, v in d.items()}
    elif isinstance(d, (list, set)):
        return [clean_dict(v) for v in d]
    else:
        return d


def recursive_remove_nones(obj: dict) -> dict:
    """
    Recursively remove None values from a dictionary.

    Args:
        obj (dict): Input dictionary.

    Returns:
        dict: Dictionary with None values removed.
    """
    _new = {}
    for k, v in obj.items():
        if isinstance(v, dict):
            _new[k] = recursive_remove_nones(v)
        elif v is not None:
            _new[k] = v
    return _new


def set_defaults(__base: dict | None = None, **kwargs) -> dict:
    """
    Set default values in a dictionary.

    Args:
        __base (dict | None, optional): Base dictionary to update. Defaults to None.
        **kwargs: Key-value pairs to set as defaults.

    Returns:
        dict: Updated dictionary with default values.
    """
    if not __base:
        __base = {}
    for k, v in kwargs.items():
        if v:
            __base[k] = v
    return __base


# List operations


def merge_lists(lists: list[list]) -> list:
    """
    Merge multiple lists, removing duplicates.

    Args:
        lists (list[list]): List of lists to merge.

    Returns:
        list: Merged list with unique elements.
    """
    merged = []
    for _list in lists:
        for elem in _list:
            if elem not in merged:
                merged.append(elem)
    return merged


def flatten_list(lst: list) -> list:
    """
    Flatten a list of lists.

    Args:
        lst (list): List to flatten.

    Returns:
        list: Flattened list.
    """
    flattened = []
    for item in lst:
        if isinstance(item, list):
            flattened.extend(flatten_list(item))
        else:
            flattened.append(item)
    return flattened


# Object conversion and manipulation


def recursive_to_dict(o: typing.Any) -> typing.Any:
    """
    Recursively convert an object to a dictionary.

    Args:
        o (typing.Any): Object to convert.

    Returns:
        typing.Any: Converted object.
    """
    if isinstance(o, dict):
        return {k: recursive_to_dict(v) for k, v in o.items()}
    elif hasattr(o, "__dict__"):
        return recursive_to_dict(o.__dict__)
    elif isinstance(o, list):
        return [recursive_to_dict(v) for v in o]
    else:
        return o


def to_dict(obj: typing.Any) -> dict:
    """
    Convert various object types to a dictionary.

    Args:
        obj (typing.Any): Object to convert.

    Returns:
        dict: Dictionary representation of the object.
    """
    if hasattr(obj, "model_export"):
        return obj.model_export()
    elif dataclasses.is_dataclass(obj):
        if isinstance(obj, type):
            # If obj is a dataclass type, not an instance
            return {
                field.name: getattr(field, "default", None)
                for field in dataclasses.fields(obj)
            }
        else:
            # If obj is a dataclass instance
            return dataclasses.asdict(obj)
    elif hasattr(obj, "json"):
        return orjson.loads(obj.json())
    elif hasattr(obj, "dict"):
        return obj.dict()
    elif hasattr(obj, "as_dict"):
        return obj.as_dict()
    elif hasattr(obj, "_asdict"):
        return obj._asdict()
    elif hasattr(obj, "to_dict"):
        return obj.to_dict()
    else:
        raise TypeError(f"Object of type {type(obj)} is not convertible to dict")


def map_object(obj: dict, **mappers) -> dict:
    """
    Apply functions to specific keys in a dictionary.

    Args:
        obj (dict): Input dictionary.
        **mappers: Functions to apply to specific keys.

    Returns:
        dict: Dictionary with mapped values.
    """
    _new = {}
    for k, v in obj.items():
        if k in mappers and v:
            if callable(mappers[k]):
                _v = mappers[k](v)
                if _v:
                    _new[k] = _v
                else:
                    _new[k] = v
            else:
                _new[k] = v
        else:
            _new[k] = v
    return _new


# JSON and file operations


def get_records_from_glob(
    path: str, limit: int | None = None, filters: list[typing.Callable] = []
) -> list[dict]:
    """
    Get records from JSON files matching a glob pattern.

    Args:
        path (str): Glob pattern for JSON files.
        limit (int | None, optional): Maximum number of records to return. Defaults to None.
        filters (list[typing.Callable], optional): List of filter functions to apply. Defaults to [].

    Returns:
        list[dict]: List of records from JSON files.
    """
    records = []
    for file in glob.glob(path):
        for record in jsonlines.open(file, loads=orjson.loads):
            if all([f(record) for f in filters]) or len(filters) == 0:
                records.append(record)
            if limit and len(records) >= limit:
                return records

    return records


def get_data_from_records(
    records: list[dict], path: str
) -> typing.Generator[typing.Any, typing.Any, None]:
    """
    Get nested data from a list of records using a dot-separated path.

    Args:
        records (list[dict]): List of records to search.
        path (str): Dot-separated path to the desired data.

    Yields:
        typing.Any: Nested data from records.
    """
    for record in records:
        o = recursive_get(record, *path.split("."))
        if o:
            yield o


def expand_nested_json(obj: dict) -> dict:
    """
    Expand nested JSON strings within a dictionary.

    Args:
        obj (dict): Dictionary containing nested JSON strings.

    Returns:
        dict: Dictionary with expanded nested JSON.

    Raises:
        ValueError: If the expanded object is not a dictionary.
    """

    def _expand_nested_json(obj) -> dict | list:
        if isinstance(obj, dict):
            for k, v in obj.items():
                if isinstance(v, str):
                    try:
                        _obj = orjson.loads(v)
                        obj[k] = _expand_nested_json(_obj)
                    except ValueError:
                        obj[k] = v
                else:
                    obj[k] = _expand_nested_json(v)
        elif isinstance(obj, list):
            for i, v in enumerate(obj):
                if isinstance(v, str):
                    try:
                        _obj = orjson.loads(v)
                        obj[i] = _expand_nested_json(_obj)
                    except ValueError:
                        obj[i] = v
                else:
                    obj[i] = _expand_nested_json(v)
        return obj

    _obj = _expand_nested_json(obj)
    if isinstance(_obj, dict):
        return _obj
    else:
        raise ValueError(f"Expanded object {_obj=} not a dictionary")


# Nested data access and manipulation


def recursive_default(obj: dict, *keys) -> typing.Any:
    """
    Get a value from nested dictionaries using multiple keys.

    Args:
        obj (dict): Dictionary to search.
        *keys: Keys to search for.

    Returns:
        typing.Any: Value found or None if not found.
    """
    for i, key in enumerate(keys):
        if isinstance(obj, dict):
            if key in obj:
                obj = obj[key]
            else:
                # If it's the last key and not found, return the entire object
                return obj if i == len(keys) - 1 else None
        else:
            return None
    return obj


def recursive_get(obj: dict | None, *keys) -> typing.Any:
    """
    Get a value from nested dictionaries using multiple keys.

    Args:
        obj (dict | None): Dictionary to search.
        *keys: Keys to search for.

    Returns:
        typing.Any: Value found or None if not found.
    """
    if obj is None:
        return None
    if len(keys) == 0:
        return obj

    if isinstance(keys[0], list):
        _obj = recursive_default(obj, *keys[0])
        if isinstance(_obj, dict):
            return recursive_get(_obj, *keys[1:])
        else:
            return _obj

    if keys[0] in obj:
        return recursive_get(obj.get(keys[0]), *keys[1:])
    return None


def recursive_convert_lists(obj: dict | list | typing.Any) -> dict | list | typing.Any:
    """
    Convert nested objects with numeric string keys to lists.

    Args:
        obj (dict): Dictionary to convert.

    Returns:
        dict: Dictionary with converted nested objects.
    """
    # if isinstance(obj, dict):
    #     for k, v in obj.items():
    #         if isinstance(v, dict) and all(
    #             [isinstance(k, str) and k.isdigit() for k in v.keys()]
    #         ):
    #             obj[k] = [recursive_convert_lists(v) for v in v.values()]
    #         elif isinstance(v, dict):
    #             obj[k] = recursive_convert_lists(v)
    # return obj
    if isinstance(obj, dict):
        if all(isinstance(k, str) and k.isdigit() for k in obj.keys()):
            return [
                recursive_convert_lists(v)
                for _, v in sorted(obj.items(), key=lambda x: int(x[0]))
            ]
        return {k: recursive_convert_lists(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [recursive_convert_lists(v) for v in obj]
    return obj


# Index and deindex operations

# def index_object(obj: dict, keys_as_tuples: bool = False) -> dict:
#     """
#     Create an indexed representation of a nested object.

#     Args:
#         obj (dict): Object to index.
#         keys_as_tuples (bool, optional): Whether to use tuples as keys. Defaults to False.

#     Returns:
#         dict: Indexed representation of the object.
#     """
#     def _index_object(obj, path=None):
#         if path is None:
#             path = []
#         if isinstance(obj, dict):
#             for k, v in obj.items():
#                 yield from _index_object(v, path + [k])
#         elif isinstance(obj, list):
#             for i, v in enumerate(obj):
#                 yield from _index_object(v, path + [f'[{i}]'])
#         else:
#             yield tuple(path), obj

#     return {
#         path_tuple_to_string(path) if not keys_as_tuples else path: value
#         for path, value in _index_object(obj)
#     }


def index_object(obj: dict, keys_as_tuples: bool = False) -> dict:
    def _index_object(obj, path=None):
        if path is None:
            path = []
        if isinstance(obj, dict):
            for k, v in obj.items():
                yield from _index_object(v, path + [k])
        elif isinstance(obj, list):
            for i, v in enumerate(obj):
                if isinstance(v, (dict, list)):
                    yield from _index_object(v, path + [f"[{i}]"])
                else:
                    yield (tuple(path + [f"[{i}]"]), v)
        else:
            yield (tuple(path), obj)

    indexed = {path: value for path, value in _index_object(obj)}
    if not keys_as_tuples:
        return {path_tuple_to_string(path): value for path, value in indexed.items()}
    return indexed


# def index_object(obj: typing.Any, prefix: str = '') -> dict[str, typing.Any]:
#     if isinstance(obj, dict):
#         return {f"{prefix}{key}": index_object(value, f"{prefix}{key}.")
#                 for key, value in obj.items() if value is not None}
#     elif isinstance(obj, list):
#         return {f"{prefix}[{i}]": index_object(value, f"{prefix}[{i}].")
#                 for i, value in enumerate(obj) if value is not None}
#     else:
#         return obj

# def deindex_object(obj: dict[str, TVal]) -> TReturn:
#     """
#     Convert an indexed object back to its nested form.

#     Args:
#         obj (dict[str, TVal]): Indexed object to convert.


#     Returns:
#         dict[str, typing.Any]: Nested form of the object.
#     """
#     # return _deindex_object(
#     #     {path_string_to_tuple(key): value for key, value in obj.items()}
#     # )
#     return _deindex_object(
#         {path_string_to_tuple(key): value for key, value in obj.items()}
#     )
def deindex_object(obj: dict[str, TVal]) -> TReturn:
    def _set_nested(result, keys, value):
        current = result
        for i, key in enumerate(keys[:-1]):
            if key.startswith("[") and key.endswith("]"):
                index = int(key[1:-1])
                if not isinstance(current, dict):
                    current = {}
                    if i == 0:
                        result = current
                    else:
                        result[keys[i - 1]] = current
                if str(index) not in current:
                    current[str(index)] = {}
                current = current[str(index)]
            else:
                if key not in current:
                    current[key] = {}
                current = current[key]

        last_key = keys[-1]
        if last_key.startswith("[") and last_key.endswith("]"):
            index = int(last_key[1:-1])
            current[str(index)] = value
        else:
            current[last_key] = value
        return result

    def _dict_to_list(d):
        if isinstance(d, dict):
            if all(k.isdigit() for k in d.keys()):
                return [_dict_to_list(d[str(i)]) for i in range(len(d))]
            return {k: _dict_to_list(v) for k, v in d.items()}
        return d

    result = {}
    for path, value in obj.items():
        keys = path.replace("]", "").replace("[", ".").split(".")
        result = _set_nested(result, keys, value)

    return _dict_to_list(result)


def index_objects(objs: list) -> list[dict]:
    """
    Create indexed representations of a list of nested objects.

    Args:
        objs (list): List of objects to index.

    Returns:
        list[dict]: List of indexed representations of the objects.
    """
    return [index_object(obj) for obj in objs]


def deindex_objects(objs: list) -> list[TReturn]:
    """
    Convert a list of indexed objects back to their nested forms.

    Args:
        objs (list): List of indexed objects to convert.

    Returns:
        list[dict]: List of nested forms of the objects.
    """
    return [deindex_object(obj) for obj in objs]


# Helper functions for index/deindex operations


def _handle_serialized_key(key: str) -> str:
    """
    Handle serialized keys in indexed objects.

    Args:
        key (str): Key to handle.

    Returns:
        str: Processed key.
    """
    if "{" in key:
        try:
            d = orjson.loads(key)
            _key = ""
            for k, v in index_object(d).items():
                _key += f"{k}_{v}__"
            return _key
        except Exception:
            return slugify.slugify(key, separator="_")
    else:
        return key


def path_tuple_to_string(path: tuple) -> str:
    """
    Convert a path tuple to a string representation.

    Args:
        path (tuple): Path tuple to convert.

    Returns:
        str: String representation of the path.
    """
    _path = ".".join(
        f"[{i}]" if isinstance(i, int) else _handle_serialized_key(i) for i in path
    )
    return _path.replace(".[", "[")


def path_string_to_tuple(path: str) -> tuple:
    """
    Convert a string path to a tuple representation.

    Args:
        path (str): String path to convert.

    Returns:
        tuple: Tuple representation of the path.
    """
    # return tuple(int(i[1:-1]) if i.startswith("[") else i for i in path.split("."))
    return tuple(
        int(i[1:-1]) if i.startswith("[") and i.endswith("]") else i
        for i in path.split(".")
    )


TReturn = dict[TKey, typing.Any] | list[typing.Any] | str | int | float | bool | None

# def _deindex_object(obj: dict[tuple[TKey, ...], TVal]) -> TReturn:
#     """
#     Convert an indexed object back to its nested form.

#     Args:
#         obj (dict[tuple[TKey, ...], TVal]): Indexed object to convert.

#     Returns:
#         TReturn: Nested form of the object, which can be:
#             - A dictionary with string or int keys and any values
#             - A list of any values
#             - A primitive value (str, int, float, bool, or None)
#     """
#     result: dict[TKey, typing.Any] = {}
#     for key, value in obj.items():
#         _set_nested_dict(result, key, value)
#     return _numeric_dicts_to_list(result)


def _deindex_object(obj: dict[tuple[TKey, ...], TVal]) -> TReturn:
    result: dict[TKey, typing.Any] = {}
    for key, value in obj.items():
        current = result
        for i, k in enumerate(key[:-1]):
            if isinstance(k, str) and k.startswith("[") and k.endswith("]"):
                # This is a list index
                parent_key = key[i - 1]
                if parent_key not in current:
                    current[parent_key] = []
                index = int(k[1:-1])
                while len(current[parent_key]) <= index:
                    current[parent_key].append({})
                current = current[parent_key][index]
            else:
                if k not in current:
                    current[k] = {}
                current = current[k]
        last_key = key[-1]
        if (
            isinstance(last_key, str)
            and last_key.startswith("[")
            and last_key.endswith("]")
        ):
            parent_key = key[-2]
            index = int(last_key[1:-1])
            if parent_key not in current:
                current[parent_key] = []
            while len(current[parent_key]) <= index:
                current[parent_key].append(None)
            current[parent_key][index] = value
        else:
            current[last_key] = value
    return result


def _set_nested_dict(
    d: dict[TKey, typing.Any], keys: tuple[TKey, ...], value: typing.Any
) -> None:
    """
    Set a value in a nested dictionary using a tuple of keys.

    Args:
        d (dict[TKey, typing.Any]): Dictionary to modify.
        keys (tuple[TKey, ...]): Tuple of keys representing the path.
        value (typing.Any): Value to set.

    Raises:
        TypeError: If a non-integer key is used with a list.
    """
    for key in keys[:-1]:
        if isinstance(d, list):
            if not isinstance(key, int):
                raise TypeError(
                    f"List indices must be integers, not {type(key).__name__}"
                )
            while len(d) <= key:
                d.append({})
            if not isinstance(d[key], (dict, list)):
                d[key] = {}
            d = d[key]
        else:
            d = d.setdefault(key, {})

    if isinstance(d, list):
        if not isinstance(keys[-1], int):
            raise TypeError(
                f"List indices must be integers, not {type(keys[-1]).__name__}"
            )
        while len(d) <= keys[-1]:
            d.append({})
        d[keys[-1]] = value
    else:
        d[keys[-1]] = value


def _numeric_dicts_to_list(d: dict[TKey, TVal] | TVal) -> TReturn:
    """
    Convert numeric dictionaries to lists recursively.

    Args:
        d (dict[TKey, TVal] | TVal): Dictionary or value to convert.

    Returns:
        TReturn: Converted structure, which can be:
            - A dictionary with string or int keys and any values
            - A list of any values
            - A primitive value (str, int, float, bool, or None)
    """
    if isinstance(d, dict):
        if all(isinstance(k, str) and k.isdigit() for k in d.keys()):
            return [_numeric_dicts_to_list(d[str(i)]) for i in range(len(d))]
        return {k: _numeric_dicts_to_list(v) for k, v in d.items()}
    elif isinstance(d, list):
        return [_numeric_dicts_to_list(item) for item in d]
    else:
        return d


# Tree-like defaultdict
# ruff:  noqa: E731
tree = lambda: collections.defaultdict(tree)


def recursive_defaultdict_to_dict(obj: collections.defaultdict) -> dict:
    """
    Convert a recursive defaultdict to a regular dictionary.

    Args:
        obj (collections.defaultdict): Recursive defaultdict to convert.

    Returns:
        dict: Regular dictionary representation.
    """
    if isinstance(obj, collections.defaultdict):
        return {k: recursive_defaultdict_to_dict(v) for k, v in obj.items()}
    else:
        return obj


def convert_index_object_to_dict(index):
    result = tree()
    for path, val in index:
        d = result
        for element in path:
            d = d[element]
        d = val
    return result


# DotDict implementation


class DotDict(MutableMapping):
    def __init__(self, init_dict: DictLike | None = None, **kwargs: typing.Any):
        self.__dict__["_data"] = {}
        if init_dict:
            self.update(init_dict)
        self.update(kwargs)

    def __getattr__(self, key: str) -> typing.Any:
        return self[key]

    def __setattr__(self, key: str, value: typing.Any) -> None:
        self[key] = value

    def __getitem__(self, key: str) -> typing.Any:
        return self._data[key]

    def __setitem__(self, key: str, value: typing.Any) -> None:
        self._data[key] = self._convert(value)

    def __delitem__(self, key: str) -> None:
        del self._data[key]

    def __iter__(self) -> typing.Iterator[str]:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)

    def __repr__(self) -> str:
        return f"DotDict({self._data})"

    def _convert(self, value: typing.Any) -> typing.Any:
        if isinstance(value, dict) and not isinstance(value, DotDict):
            return DotDict(value)
        elif isinstance(value, list):
            return [self._convert(v) for v in value]
        return value

    def update(self, other: DictLike | None = None, **kwargs: typing.Any) -> None:
        if other is not None:
            for k, v in other.items():
                self[k] = v
        for k, v in kwargs.items():
            self[k] = v

    def to_dict(self) -> dict:
        def _to_dict(value):
            if isinstance(value, DotDict):
                return value.to_dict()
            elif isinstance(value, list):
                return [_to_dict(v) for v in value]
            return value

        return {k: _to_dict(v) for k, v in self._data.items()}


# class DotDict(MutableMapping):
#     """
#     A dictionary that supports attribute-style access (dot notation).

#     This class extends the standard python dictionary to allow accessing
#     and setting items using dot notation (e.g., `d.key` instead of `d['key']`).

#     Args:
#         init_dict (DictLike | None, optional): Initial dictionary to populate the DotDict. Defaults to None.
#         **kwargs: Additional key-value pairs to add to the DotDict.

#     Example:
#         >>> d = DotDict({'a': 1, 'b': {'c': 2}})
#         >>> d.a
#         1
#         >>> d.b.c
#         2
#         >>> d.d = 3
#         >>> d['d']
#         3
#     """

#     def __init__(self, init_dict: DictLike | None = None, **kwargs: typing.Any):
#         if init_dict:
#             super().update(init_dict)
#         super().update(kwargs)

#     def __getitem__(self, key: str) -> typing.Any:
#         return self.__dict__[key]

#     def __setitem__(self, key: str, value: typing.Any) -> None:
#         self.__dict__[key] = value

#     def __delitem__(self, key: str) -> None:
#         del self.__dict__[key]

#     def __iter__(self) -> typing.Iterator[str]:
#         return iter(self.__dict__)

#     def __len__(self) -> int:
#         return len(self.__dict__)

#     def __repr__(self) -> str:
#         if len(self) > 0:
#             return f"<{type(self).__name__}: {', '.join(sorted(repr(k) for k in self.keys()))}>"
#         else:
#             return f"<{type(self).__name__}>"

#     def get(self, key: str, default: typing.Any = None) -> typing.Any:
#         """
#         Get a value from the DotDict.

#         Args:
#             key (str): The key to retrieve.
#             default (typing.Any, optional): Default value to return if the key is not found. Defaults to None.

#         Returns:
#             typing.Any: The value associated with the key, or the default value if not found.
#         """
#         return super().get(key, default)

#     def copy(self) -> "DotDict":
#         """
#         Create a shallow copy of the DotDict.

#         Returns:
#             DotDict: A new DotDict instance with copied key-value pairs.
#         """
#         return type(self)(self.__dict__.copy())

#     def to_dict(self) -> dict:
#         """
#         Convert the DotDict to a regular dictionary.

#         Returns:
#             dict: A regular dictionary representation of the DotDict.
#         """
#         return typing.cast(dict, as_nested_dict(self, dct_class=dict))

# Helper functions for DotDict


def is_iterable(obj: typing.Any) -> typing.TypeGuard[Iterable]:
    """
    Check if an object is iterable (but not a string).

    Args:
        obj (typing.Any): Object to check.

    Returns:
        bool: True if the object is iterable and not a string, False otherwise.
    """
    return isinstance(obj, Iterable) and not isinstance(obj, (str, bytes))


def as_nested_dict(
    obj: MutableMapping[typing.Any, typing.Any] | Iterable[typing.Any],
    dct_class: type[MutableMapping] | typing.Callable[[], MutableMapping] = dict,
) -> MutableMapping[typing.Any, typing.Any] | typing.Any:
    """
    Convert a nested structure to the specified dictionary class.

    Args:
        obj (MutableMapping[typing.Any, typing.Any] | Iterable[typing.Any]): Object to convert.
        dct_class (type[MutableMapping] | Callable[[], MutableMapping], optional):
            Dictionary class or function that returns a new dictionary instance. Defaults to dict.

    Returns:
        MutableMapping[typing.Any, typing.Any] | typing.Any: Converted structure.
    """
    if isinstance(obj, (list, tuple, set)):
        return type(obj)(as_nested_dict(d, dct_class) for d in obj)
    elif isinstance(obj, Box):
        return dict(obj)
    elif isinstance(obj, MutableMapping):
        new_dict = dct_class() if callable(dct_class) else dct_class()
        for k, v in getattr(obj, "__dict__", obj).items():
            new_dict[k] = as_nested_dict(v, dct_class)
        return new_dict
    return obj


# Compound key operations


class CompoundKey(tuple):
    """A tuple subclass used for representing nested dictionary keys."""

    pass


def dict_to_flatdict(dct: DictLike, parent: CompoundKey | None = None) -> dict:
    """
    Convert a nested dictionary to a flat dictionary with compound keys.

    Args:
        dct (DictLike): Nested dictionary to flatten.
        parent (CompoundKey | None, optional): Parent key for recursive calls. Defaults to None.

    Returns:
        dict: Flattened dictionary with compound keys.
    """
    items = []
    parent = parent or CompoundKey()
    for k, v in dct.items():
        k_parent = CompoundKey(parent + (k,))
        if isinstance(v, dict):
            items.extend(dict_to_flatdict(v, parent=k_parent).items())
        else:
            items.append((k_parent, v))
    return dict(items)


def flatdict_to_dict(dct: dict, dct_class: typing.Type[D] | None = None) -> D:
    """
    Convert a flat dictionary with compound keys back to a nested dictionary.

    Args:
        dct (dict): Flat dictionary with compound keys.
        dct_class (typing.Type[D] | None, optional): Dictionary class to use for the result. Defaults to None.

    Returns:
        D: Nested dictionary of the specified class.
    """
    result = typing.cast(D, (dct_class or dict)())
    for k, v in dct.items():
        if isinstance(k, CompoundKey):
            current_dict = result
            for ki in k[:-1]:
                current_dict = current_dict.setdefault(  # type: ignore
                    ki, (dct_class or dict)()
                )
            current_dict[k[-1]] = v
        else:
            result[k] = v

    return result


# Utility function for flattening sequences
def flatten_seq(seq: Iterable) -> typing.Generator:
    """
    Flatten a nested sequence (e.g., list of lists) into a single level.

    Args:
        seq (Iterable): Nested sequence to flatten.

    Yields:
        typing.Any: Elements from the flattened sequence.

    Example:
        >>> list(flatten_seq([1, [2, 3, [4, 5]], 6]))
        [1, 2, 3, 4, 5, 6]
    """
    for item in seq:
        if isinstance(item, Iterable) and not isinstance(item, (str, bytes)):
            yield from flatten_seq(item)
        else:
            yield item



import pytest
from good_common.utilities import (
    try_chain,
    deep_attribute_get,
    deep_attribute_set,
    set_defaults,
    filter_nulls,
)


# Test for try_chain
def test_try_chain():
    def fail_func():
        raise ValueError("Failed")

    def success_func(x):
        return x * 2

    chain = try_chain([fail_func, success_func])
    assert chain(5) == 10

    fail_chain = try_chain([fail_func, fail_func], fail=True)
    with pytest.raises(ValueError):
        fail_chain(5)

    default_chain = try_chain([fail_func, fail_func], default_value="default")
    assert default_chain(5) == "default"


# Tests for deep_attribute_get
def test_deep_attribute_get():
    obj = {
        "a": {"b": [{"c": 1}, {"c": 2}], "d": {"e": "value"}},
        "f": [1, 2, 3],
        "g": {"h": [{"i": "test"}]},
    }

    assert deep_attribute_get(obj, "a.b[0].c") == 1
    assert deep_attribute_get(obj, "a.b[*].c") == [1, 2]
    assert deep_attribute_get(obj, "a.d.e") == "value"
    assert deep_attribute_get(obj, "f[1]") == 2
    assert deep_attribute_get(obj, "g.h[0].i") == "test"
    assert deep_attribute_get(obj, "nonexistent", default="not found") == "not found"

    result = deep_attribute_get(obj, "a.b[*].c", return_paths=True)
    assert result == [(1, "a.b[0].c"), (2, "a.b[1].c")]


# Test for deep_attribute_set
def test_deep_attribute_set():
    obj = {"a": {"b": {"c": 1}}}
    deep_attribute_set(obj, "a.b.c", 2)
    assert obj["a"]["b"]["c"] == 2

    deep_attribute_set(obj, "a.b.d", 3)
    assert obj["a"]["b"]["d"] == 3

    deep_attribute_set(obj, "x.y.z", 4)
    assert obj["x"]["y"]["z"] == 4


# Test for set_defaults
def test_set_defaults():
    base = {"a": 1, "b": None}
    result = set_defaults(base, b=2, c=3)
    assert result == {"a": 1, "b": 2, "c": 3}

    result = set_defaults(b=2, c=3)
    assert result == {"b": 2, "c": 3}


# Test for filter_nulls
def test_filter_nulls():
    obj = {
        "a": 1,
        "b": None,
        "c": {"d": 2, "e": None, "f": [1, None, 3]},
        "g": [{"h": 4, "i": None}, {"j": None}, {"k": 5}],
    }

    result = filter_nulls(obj)
    assert result == {"a": 1, "c": {"d": 2, "f": [1, 3]}, "g": [{"h": 4}, {"k": 5}]}


# Additional tests for edge cases and specific behaviors


def test_deep_attribute_get_with_regex():
    obj = {"abc": 1, "abd": 2, "bcd": 3}
    assert deep_attribute_get(obj, "ab.") == [1, 2]


def test_deep_attribute_get_with_nonexistent_path():
    obj = {"a": {"b": 1}}
    assert deep_attribute_get(obj, "a.c.d", default="not found") == "not found"


def test_try_chain_with_mixed_functions():
    def int_func(x):
        return int(x)

    def float_func(x):
        return float(x)

    def str_func(x):
        return str(x)

    chain = try_chain([int_func, float_func, str_func])
    assert chain("10") == 10
    assert chain("10.5") == 10.5
    assert chain("abc") == "abc"


def test_set_defaults_with_falsy_values():
    base = {"a": 0, "b": ""}
    result = set_defaults(base, a=1, b="test", c=False)
    assert result == {"a": 0, "b": "", "c": False}


def test_filter_nulls_with_empty_containers():
    obj = {"a": [], "b": {}, "c": [None], "d": {"e": None}}
    result = filter_nulls(obj)
    assert result == {}



from ._asyncio import run_async, async_iterator

from ._orchestration import parse_args, name_process, KeyboardInterruptHandler
from ._functional import (
    try_chain,
    anyget,
    anyset,
    deep_attribute_get,
    deep_attribute_set,
    set_defaults,
    filter_nulls,
)
from ._dates import parse_timestamp, parse_date

from ._data import (
    is_int,
    to_int,
    signed_64_to_unsigned_128,
    int_to_base62,
    to_float,
    to_numeric,
    compressed_farmhash,
    farmhash_compressed,
    farmhash_bytes,
    farmhash_hex,
    b64_decode,
    b64_encode,
)

from ._collections import (
    sort_object_keys,
    object_farmhash,
    map_object,
    recursive_convert_lists,
    recursive_remove_nones,
    expand_nested_json,
    index_object,
    convert_index_object_to_dict,
    deindex_object,
    index_objects,
    deindex_objects,
)

from ._iterators import iter_ranges

from ._logging import catchtime

from ._strings import (
    is_uuid,
    camel_to_slug,
    camel_to_snake,
    camel_to_kebab,
    snake_to_camel,
    snake_to_kebab,
    kebab_to_camel,
    kebab_to_snake,
    slug_to_camel,
    slug_to_snake,
    slug_to_kebab,
    detect_string_type
)

from ._yaml import yaml_load, yaml_dump, yaml_loads, yaml_dumps


__all__ = [
    "anyget",
    "anyset",
    "b64_decode",
    "b64_encode",
    "camel_to_slug",
    "camel_to_snake",
    "camel_to_kebab",
    "snake_to_camel",
    "snake_to_kebab",
    "kebab_to_camel",
    "kebab_to_snake",
    "slug_to_camel",
    "slug_to_snake",
    "slug_to_kebab",
    "catchtime",
    "compressed_farmhash",
    "convert_index_object_to_dict",
    "deep_attribute_get",
    "deep_attribute_set",
    "deindex_object",
    "deindex_objects",
    "detect_string_type",
    "expand_nested_json",
    "farmhash_bytes",
    "farmhash_compressed",
    "farmhash_hex",
    "filter_nulls",
    "index_object",
    "index_objects",
    "int_to_base62",
    "is_int",
    "is_uuid",
    "iter_ranges",
    "KeyboardInterruptHandler",
    "sort_object_keys",
    "object_farmhash",
    "map_object",
    "name_process",
    "parse_args",
    "parse_date",
    "parse_timestamp",
    "recursive_convert_lists",
    "recursive_remove_nones",
    "run_async",
    "async_iterator",
    "set_defaults",
    "signed_64_to_unsigned_128",
    "to_float",
    "to_int",
    "to_numeric",
    "try_chain",
    "yaml_dump",
    "yaml_dumps",
    "yaml_load",
    "yaml_loads",
]



import datetime
import typing

from dateutil import parser

from ._functional import try_chain


def strptime(format: str) -> typing.Callable[..., datetime.datetime]:
    def _strptime(value):
        return datetime.datetime.strptime(value, format)

    return _strptime


def parse_timestamp(
    timestamp: typing.Any,
    *formats,
    raise_error=False,
    timezone=datetime.UTC,
    as_date=False,
) -> typing.Optional[datetime.datetime]:
    if not timestamp:
        return None

    if isinstance(timestamp, datetime.datetime):
        return timestamp

    chain_fns = []

    if len(formats) > 0:
        for fmt in formats:
            chain_fns.append(strptime(fmt))

    chain_fns += [
        strptime("%Y-%m-%dT%H:%M:%S.%fZ"),
        strptime("%Y-%m-%d"),
        lambda x: parser.parse(x),
    ]

    output = try_chain(chain_fns, fail=raise_error)(timestamp)
    if output:
        output.replace(tzinfo=timezone)
        return output.date() if as_date else output
    return None


def parse_date(
    date: typing.Any, *formats, raise_error=False
) -> typing.Optional[datetime.date]:
    return parse_timestamp(
        date,
        *formats,
        raise_error=raise_error,
        as_date=True,
    )
